/*++

Copyright (c) 1996  Microsoft Corporation

Module Name:

    filter.h

Abstract: NULL filter driver -- boilerplate code

Author:

    ervinp

Environment:

    Kernel mode

Revision History:


--*/

enum deviceState {
        STATE_INITIALIZED,
        STATE_STARTING,
        STATE_STARTED,
        STATE_START_FAILED,
        STATE_STOPPED,  // implies device was previously started successfully
        STATE_SUSPENDED,
        STATE_REMOVING,
        STATE_REMOVED
};


/*
 *  Memory tag for memory blocks allocated by this driver
 *  (used in ExAllocatePoolWithTag() call).
 *  This DWORD appears as "Filt" in a little-endian memory byte dump.
 *
 *  NOTE:  PLEASE change this value to be unique for your driver!  Otherwise,
 *  your allocations will show up with every other driver that uses 'tliF' as
 *  an allocation tag.
 *  
 */
#define FILTER_TAG (ULONG)'V63X'

#undef ExAllocatePool
#define ExAllocatePool(type, size) \
            ExAllocatePoolWithTag (type, size, FILTER_TAG)


// X36 configuration data structures.

typedef struct _SCRIPT {
	unsigned char	Op; 	// SET or RESET, keyb or button
	unsigned char	Offset; // Bitpos in table.
	unsigned long	Delay;  // Absolute delay in ms.
	struct _SCRIPT *Next;   // Next entry.
} SCRIPT, *PSCRIPT;

typedef struct _AXIS {
	unsigned int 	Max;
	unsigned int 	Min;
	// Control data will go here.
} AXIS, *PAXIS;

typedef struct _TWAY {
	PSCRIPT			Pressed;
	PSCRIPT			Released;
	PSCRIPT			Hold;
} TWAY, *PTWAY;

typedef struct _TRIGGER {
	PSCRIPT 		Pressed;
	PSCRIPT			Released;
	PSCRIPT			Hold;
} TRIGGER, *PTRIGGER;

typedef struct _HATSW {
	unsigned char	Mode;
	PSCRIPT			Pressed[9];
	PSCRIPT			Released[9];
	PSCRIPT			Hold[9];
} HATSW, *PHATSW;

typedef struct _CONFIG {
	AXIS	Axis[6];
	TWAY	Slider[2];
	TRIGGER	Button[8];
	HATSW	Hat[4];
} CONFIG, *PCONFIG;


// X36 queue data structures.
typedef struct _QENTRY {
	unsigned char	Op; 	// SET or RESET, keyb or button
	unsigned char	Offset; // Bitpos in table.
	LARGE_INTEGER	Time;   // System time it goes off.
	struct _QENTRY *Next;   // Next entry.
} QENTRY, *PQENTRY;
	

typedef struct _XQUEUE {
	PQENTRY Head;
	KSPIN_LOCK Lock;
} XQUEUE, *PXQUEUE;



// Driver extension.

#define DEVICE_EXTENSION_SIGNATURE 'V63X'

typedef struct DEVICE_EXTENSION {

    /*
     *  Memory signature of a device extension, for debugging.
     */
    ULONG signature;

    /*
     *  Plug-and-play state of this device object.
     */
    enum deviceState state;

    /*
     *  The device object that this filter driver created.
     */
    PDEVICE_OBJECT filterDevObj;

    /*
     *  The device object created by the next lower driver.
     */
    PDEVICE_OBJECT physicalDevObj;

    /*
     *  The device object at the top of the stack that we attached to.
     *  This is often (but not always) the same as physicalDevObj.
     */
    PDEVICE_OBJECT topDevObj;

    /*
     *  deviceCapabilities includes a
     *  table mapping system power states to device power states.
     */
    DEVICE_CAPABILITIES deviceCapabilities;

    /*
     *  pendingActionCount is used to keep track of outstanding actions.
     *  removeEvent is used to wait until all pending actions are
     *  completed before complete the REMOVE_DEVICE IRP and let the
     *  driver get unloaded.
     */
    LONG pendingActionCount;
    KEVENT removeEvent;


	
	// X36 related stuffs.
	PIRP ReadIrp;  // Read request IRP, held for keyboard replies.
	KSPIN_LOCK						ReadLock; // Spinlock on event buffer.
	KDPC							TimerDPC; // For timer.
	KDPC							ReturnDPC; // To fill out and return packets.
	KTIMER							Timer; // For scripting.
	unsigned char					NewData; // Presence of data to return to user.
	
	LARGE_INTEGER					StartTime; // Check packet delay.
	
	XQUEUE ActionQueue; // Timer queue with actions to perform.
	
	// X36 state data.
	unsigned char LastPosition[0x0b];  // Previous joystick packet, raw form.
	unsigned char KeyboardStatus[29]; // DirectX buttons are embedded here too.
	PCONFIG		  Configurations[4][4][3]; 	 // Configuration matrix.  Pointers, so we can alloc mem dynamically.		
};




#if DBG
    #define DBGOUT(params_in_parentheses)   \
        {                                               \
            DbgPrint("'FILTER> "); \
            DbgPrint params_in_parentheses; \
            DbgPrint("\n"); \
        }
    #define TRAP(msg)  \
        {   \
            DBGOUT(("TRAP at file %s, line %d: '%s'.", __FILE__, __LINE__, msg)); \
            DbgBreakPoint(); \
        }
#else
    #define DBGOUT(params_in_parentheses)
    #define TRAP(msg)
#endif


/*
 *  Function externs
 */
NTSTATUS    DriverEntry(IN PDRIVER_OBJECT DriverObject, IN PUNICODE_STRING RegistryPath);
NTSTATUS    VA_AddDevice(IN PDRIVER_OBJECT driverObj, IN PDEVICE_OBJECT pdo);
VOID        VA_DriverUnload(IN PDRIVER_OBJECT DriverObject);
NTSTATUS    VA_Dispatch(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);
NTSTATUS    VA_PnP(struct DEVICE_EXTENSION *devExt, PIRP irp);
NTSTATUS    VA_Power(struct DEVICE_EXTENSION *devExt, PIRP irp);
NTSTATUS    VA_PowerComplete(IN PDEVICE_OBJECT devObj, IN PIRP irp, IN PVOID context);
NTSTATUS    GetDeviceCapabilities(struct DEVICE_EXTENSION *devExt);
NTSTATUS    CallNextDriverSync(struct DEVICE_EXTENSION *devExt, PIRP irp);
NTSTATUS    CallDriverSync(PDEVICE_OBJECT devObj, PIRP irp);
NTSTATUS    CallDriverSyncCompletion(IN PDEVICE_OBJECT devObj, IN PIRP irp, IN PVOID Context);
VOID        IncrementPendingActionCount(struct DEVICE_EXTENSION *devExt);
VOID        DecrementPendingActionCount(struct DEVICE_EXTENSION *devExt);
NTSTATUS    QueryDeviceKey(HANDLE Handle, PWCHAR ValueNameString, PVOID Data, ULONG DataLength);
VOID        RegistryAccessSample(struct DEVICE_EXTENSION *devExt, PDEVICE_OBJECT devObj);

// IOCTL.C
NTSTATUS    VA_Ioctl(struct DEVICE_EXTENSION *devExt, PIRP Irp);
NTSTATUS IoCtlCompletion( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context );
VOID ExecuteScript( IN PKDPC Dpc, IN PVOID DeferredContext, IN PVOID SystemArgument1, IN PVOID SystemArgument2);
void CompleteRead( IN PKDPC Dpc, IN PVOID DeferredContext, IN PVOID SystemArgument1, IN PVOID SystemArgument2);


// profile.C
NTSTATUS LoadProfile(struct DEVICE_EXTENSION  *DeviceExtension);
NTSTATUS ClearProfile(struct DEVICE_EXTENSION  *DeviceExtension);
NTSTATUS ReloadProfile(struct DEVICE_EXTENSION  *DeviceExtension);

// Queue.c
NTSTATUS QueueInit(PXQUEUE MyQueue);
NTSTATUS AddScript(PXQUEUE MyQueue, PSCRIPT MyScript, LARGE_INTEGER RefTime);
PQENTRY QueuePop(PXQUEUE MyQueue); // USER MUST ExFreePool QUEUE ENTRY!
PQENTRY QueuePeek(PXQUEUE MyQueue);
void QueueClear(PXQUEUE MyQueue);
void AddAllScripts(struct DEVICE_EXTENSION *devExt, unsigned char mode, unsigned char aux, unsigned char shift, void* Config0Pos, LARGE_INTEGER Time);

