// QUEUE.C
//
// Handles the timer queue for the X36 scripting engine.
// 
// includes:
//
// queueinit: inits queue.
// scriptadd: adds a whole script to the queue.
// pop: removes the first item from the queue.
// peek: returns the first item in the queue without destroying it
// queueclear: destroys a queue.

// #define DBG 1
#include <WDM.H>
#include "filter.h"


NTSTATUS QueueInit(PXQUEUE MyQueue)
{
	if (MyQueue == NULL)
		return STATUS_INVALID_PARAMETER;
	
	MyQueue->Head = NULL;
	KeInitializeSpinLock(&MyQueue->Lock);
	
	return STATUS_SUCCESS;
}

NTSTATUS AddScript(PXQUEUE MyQueue, PSCRIPT MyScript, LARGE_INTEGER RefTime)
{
	PSCRIPT SCurrent;
	PQENTRY QCurrent, QPrev, QNew;
	KIRQL OldIrqL;
	
	if (MyQueue == NULL)
		return STATUS_INVALID_PARAMETER;
		
	KeAcquireSpinLock(&MyQueue->Lock, &OldIrqL);

	QPrev = NULL;
	QCurrent = MyQueue->Head;
	SCurrent = MyScript;
	
	while (SCurrent != NULL)
	{
		if (QNew = ExAllocatePool(PagedPool, sizeof(QENTRY)))
		{
			RtlZeroMemory(QNew,sizeof(QENTRY));
			QNew->Op = SCurrent->Op;
			QNew->Offset = SCurrent->Offset;
			QNew->Time = RtlLargeIntegerAdd(RtlConvertLongToLargeInteger(SCurrent->Delay),RefTime);
//			DbgPrint("Queuing item:   Op: %d   Offset: %d   Time: %x %x\n",QNew->Op,QNew->Offset,QNew->Time.HighPart,QNew->Time.LowPart);
			
			if (QCurrent == NULL)
				MyQueue->Head = QCurrent = QNew;
			else
			{
				while((QCurrent != NULL) && (RtlLargeIntegerGreaterThan(QNew->Time,QCurrent->Time)))
				{
					QPrev = QCurrent;
					QCurrent = QCurrent -> Next;
				}
				
				
				if (QPrev != NULL)
				{
					QNew->Next = QCurrent;
					QPrev->Next = QNew;
				}
				else
				{
					MyQueue->Head = QNew;
					QNew->Next = QCurrent;				
				}
				QCurrent=QNew;
			}
			SCurrent= SCurrent->Next;			
		}
		else
		{
			KeReleaseSpinLock(&MyQueue->Lock, OldIrqL);
			return STATUS_INSUFFICIENT_RESOURCES;
		}
	}
		
	KeReleaseSpinLock(&MyQueue->Lock, OldIrqL);
	return STATUS_SUCCESS;
}



PQENTRY QueuePop(PXQUEUE MyQueue) // USER MUST FREE QUEUE ENTRY!
{
	PQENTRY QCurrent;
	KIRQL OldIrqL;
	
	if (MyQueue == NULL)
		return NULL;
		
	KeAcquireSpinLock(&MyQueue->Lock, &OldIrqL);
	
	QCurrent = MyQueue->Head;
	if (QCurrent != NULL)
		MyQueue->Head = QCurrent->Next;
	
	KeReleaseSpinLock(&MyQueue->Lock, OldIrqL);
	return QCurrent;
}

PQENTRY QueuePeek(PXQUEUE MyQueue)
{
	PQENTRY QCurrent;
	KIRQL OldIrqL;
	
	if (MyQueue == NULL)
		return NULL;
		
	KeAcquireSpinLock(&MyQueue->Lock, &OldIrqL);
	
	QCurrent = MyQueue->Head;
	
	KeReleaseSpinLock(&MyQueue->Lock, OldIrqL);
	return QCurrent;
}

void QueueClear(PXQUEUE MyQueue)
{
	PQENTRY QCurrent;
	KIRQL OldIrqL;	
	
	if (MyQueue == NULL)
		return;
		
	KeAcquireSpinLock(&MyQueue->Lock, &OldIrqL);

	while(MyQueue->Head != NULL)
	{
		QCurrent = MyQueue->Head;
		MyQueue->Head = QCurrent->Next;
		ExFreePool(QCurrent);
	}
	
	KeReleaseSpinLock(&MyQueue->Lock, OldIrqL);
}


void AddAllScripts(struct DEVICE_EXTENSION *devExt, unsigned char mode, unsigned char aux, unsigned char shift, void* Config0Pos, LARGE_INTEGER Time)
{
	unsigned int Delta;
	unsigned char i,j,k;
	
	Delta = (unsigned char*)Config0Pos - (unsigned char*)devExt->Configurations[0][0][0];
//	DbgPrint("AddAllScripts delta: %d   mode: %d   aux: %d   shift: %d\n",Delta,mode,aux,shift);
//	DbgPrint("AAS Computed: %X   Actual: %X\n",(((unsigned char*)devExt->Configurations[0][0][0])+Delta),&(devExt->Configurations[0][0][0]->Button[0].Released));

	for(i=0;i<2;i++)
		for(j=0;j<2;j++)
			for(k=0;k<2;k++)
				if (devExt->Configurations[i*mode][aux*j][shift*k])
				{
//					DbgPrint("AddAllScripts adding:  delta: %d   mode: %d   aux: %d   shift: %d\n",Delta,mode*i,aux*j,shift*k);
					AddScript(&devExt->ActionQueue,*(PSCRIPT*)(((unsigned char*)devExt->Configurations[i*mode][aux*j][shift*k])+Delta), Time);
				}
}